"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
var _exportNames = {
  initialize: true,
  destroy: true,
  start: true,
  stop: true,
  activate: true,
  deactivate: true,
  synthesize: true,
  speak: true,
  classify: true,
  isInitialized: true,
  isStarted: true,
  isActivated: true,
  addEventListener: true,
  removeEventListener: true,
  removeAllListeners: true
};
exports.default = exports.removeAllListeners = exports.removeEventListener = exports.addEventListener = exports.isActivated = exports.isStarted = exports.isInitialized = exports.classify = exports.speak = exports.synthesize = exports.deactivate = exports.activate = exports.stop = exports.start = exports.destroy = exports.initialize = void 0;

var _reactNative = require("react-native");

var _types = require("./types");

Object.keys(_types).forEach(function (key) {
  if (key === "default" || key === "__esModule") return;
  if (Object.prototype.hasOwnProperty.call(_exportNames, key)) return;
  if (key in exports && exports[key] === _types[key]) return;
  Object.defineProperty(exports, key, {
    enumerable: true,
    get: function () {
      return _types[key];
    }
  });
});

var _resolveModelUrl = _interopRequireDefault(require("./resolveModelUrl"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const {
  Spokestack
} = _reactNative.NativeModules; // Event emitter

const emitter = new _reactNative.NativeEventEmitter(Spokestack);
// This is necessary to allow usage where arguments may
// be passed accidentally.
// For example,
// Spokestack.initialize(id, secret, config).then(Spokestack.start)
// The above call to start would normally throw an error.
;
['stop', 'deactivate'].forEach(method => {
  const original = Spokestack[method];

  Spokestack[method] = () => original();
}) // start (if using wakeword) and activate also require microphone permissions
;
['start', 'activate'].forEach(method => {
  const original = Spokestack[method];

  Spokestack[method] = async () => {
    // Android requires you to ask for the permission manually
    // iOS will bring up a dialog automatically
    if (_reactNative.Platform.OS === 'android' && !(await _reactNative.PermissionsAndroid.check(_reactNative.PermissionsAndroid.PERMISSIONS.RECORD_AUDIO))) {
      return Promise.reject(`Microphone permission is required to use Spokestack.
        Use PermissionsAndroid from React Native to request microphone
        permission before calling Spokestack.start() or Spokestack.activate().`);
    }

    return original();
  };
}) // Make all but one argument optional in JS-land
;
['synthesize', 'speak'].forEach(method => {
  const original = Spokestack[method];

  Spokestack[method] = (input, format = _types.TTSFormat.TEXT, voice = 'demo-male') => original(input, format, voice);
}); // Ensure method is called with proper number of args despite what's passed

const originalInit = Spokestack.initialize;

Spokestack.initialize = (id, secret, config = {}) => {
  var _config$pipeline;

  // Set refreshModels to true in development by default
  if (typeof config.refreshModels !== 'boolean') {
    config.refreshModels = __DEV__;
  }
  /**
   * Wakeword
   */
  // Resolve source objects to URLs for local downloads


  const wakewordConfig = config.wakeword || {};

  if (typeof wakewordConfig.detect === 'number') {
    wakewordConfig.detect = (0, _resolveModelUrl.default)(wakewordConfig.detect);
  }

  if (typeof wakewordConfig.encode === 'number') {
    wakewordConfig.encode = (0, _resolveModelUrl.default)(wakewordConfig.encode);
  }

  if (typeof wakewordConfig.filter === 'number') {
    wakewordConfig.filter = (0, _resolveModelUrl.default)(wakewordConfig.filter);
  } // Convert wakewords to a comma-separated list if an array


  if (Array.isArray(wakewordConfig.wakewords)) {
    wakewordConfig.wakewords = wakewordConfig.wakewords.join(',');
  }

  const hasWakeword = !!(wakewordConfig.filter && wakewordConfig.detect && wakewordConfig.encode);
  /**
   * Keyword
   */
  // Resolve source objects to URLs for local downloads

  const keywordConfig = config.keyword || {};

  if (typeof keywordConfig.detect === 'number') {
    keywordConfig.detect = (0, _resolveModelUrl.default)(keywordConfig.detect);
  }

  if (typeof keywordConfig.encode === 'number') {
    keywordConfig.encode = (0, _resolveModelUrl.default)(keywordConfig.encode);
  }

  if (typeof keywordConfig.filter === 'number') {
    keywordConfig.filter = (0, _resolveModelUrl.default)(keywordConfig.filter);
  }

  const keywordMetadataConfig = keywordConfig;

  if (typeof keywordMetadataConfig.metadata === 'number') {
    keywordMetadataConfig.metadata = (0, _resolveModelUrl.default)(keywordMetadataConfig.metadata);
  }

  const keywordClassesConfig = keywordConfig; // Convert classes to a comma-separated list if an array

  if (Array.isArray(keywordClassesConfig.classes)) {
    keywordClassesConfig.classes = keywordClassesConfig.classes.join(',');
  }

  const hasKeyword = !!(keywordConfig.filter && keywordConfig.detect && keywordConfig.encode && (keywordClassesConfig.classes || keywordMetadataConfig.metadata)); // Set default profile based on presence of wakeword or keyword config
  // Do not set a default if profile is set explicitly

  if (!((_config$pipeline = config.pipeline) !== null && _config$pipeline !== void 0 && _config$pipeline.profile)) {
    config.pipeline = config.pipeline || {};

    if (hasWakeword && !hasKeyword) {
      // If wakeword is set and keyword is not,
      // default to a profile that works with wakeword.
      config.pipeline.profile = _types.PipelineProfile.TFLITE_WAKEWORD_NATIVE_ASR;
    } else if (!hasWakeword && hasKeyword) {
      // If keyword is set, and wakeword is not,
      // default to a profile that works with keyword.
      config.pipeline.profile = _types.PipelineProfile.VAD_KEYWORD_ASR;
    } else if (hasWakeword && hasKeyword) {
      // If both keyword and wakeword are set,
      // default to a profile that works for both.
      config.pipeline.profile = _types.PipelineProfile.TFLITE_WAKEWORD_KEYWORD;
    }
  }
  /**
   * NLU
   */


  const nluConfig = config.nlu || {};

  if (typeof nluConfig.model === 'number') {
    nluConfig.model = (0, _resolveModelUrl.default)(nluConfig.model);
  }

  if (typeof nluConfig.metadata === 'number') {
    nluConfig.metadata = (0, _resolveModelUrl.default)(nluConfig.metadata);
  }

  if (typeof nluConfig.vocab === 'number') {
    nluConfig.vocab = (0, _resolveModelUrl.default)(nluConfig.vocab);
  }

  return originalInit(id, secret, config);
}; // Remove all JS listeners as well as calling
// destroy natively


const originalDestroy = Spokestack.destroy;

Spokestack.destroy = () => {
  originalDestroy(); // Remove listeners after destroy

  Spokestack.removeAllListeners();
};

Object.assign(Spokestack, {
  // Add enums as values
  PipelineProfile: _types.PipelineProfile,
  TraceLevel: _types.TraceLevel,
  TTSFormat: _types.TTSFormat,

  /**
   * Event handling
   *
   * @example
   * const { remove } = Spokestack.addEventListener('recognize', onRecognize)
   */
  addEventListener: (type, listener) => emitter.addListener(type, listener),

  /**
   * @deprecated Just like the underlying react native method, use `remove` from the return value of `addListener`.
   */
  removeEventListener: (type, listener) => emitter.removeListener(type, listener),
  removeAllListeners: () => {
    ;
    ['recognize', 'partial_recognize', 'activate', 'deactivate', 'timeout', 'play', 'error'].forEach(event => emitter.removeAllListeners(event));
  }
}); // Export types to be used as separate exports

// Allow importing these methods separately
const {
  initialize,
  destroy,
  start,
  stop,
  activate,
  deactivate,
  synthesize,
  speak,
  classify,
  isInitialized,
  isStarted,
  isActivated,
  addEventListener,
  removeEventListener,
  removeAllListeners
} = Spokestack;
exports.removeAllListeners = removeAllListeners;
exports.removeEventListener = removeEventListener;
exports.addEventListener = addEventListener;
exports.isActivated = isActivated;
exports.isStarted = isStarted;
exports.isInitialized = isInitialized;
exports.classify = classify;
exports.speak = speak;
exports.synthesize = synthesize;
exports.deactivate = deactivate;
exports.activate = activate;
exports.stop = stop;
exports.start = start;
exports.destroy = destroy;
exports.initialize = initialize;
var _default = Spokestack;
exports.default = _default;
//# sourceMappingURL=index.js.map