import { NativeEventEmitter, NativeModules, PermissionsAndroid, Platform } from 'react-native';
import { PipelineProfile, TTSFormat, TraceLevel } from './types';
import resolveModelUrl from './resolveModelUrl';
const {
  Spokestack
} = NativeModules; // Event emitter

const emitter = new NativeEventEmitter(Spokestack);
// This is necessary to allow usage where arguments may
// be passed accidentally.
// For example,
// Spokestack.initialize(id, secret, config).then(Spokestack.start)
// The above call to start would normally throw an error.
;
['stop', 'deactivate'].forEach(method => {
  const original = Spokestack[method];

  Spokestack[method] = () => original();
}) // start (if using wakeword) and activate also require microphone permissions
;
['start', 'activate'].forEach(method => {
  const original = Spokestack[method];

  Spokestack[method] = async () => {
    // Android requires you to ask for the permission manually
    // iOS will bring up a dialog automatically
    if (Platform.OS === 'android' && !(await PermissionsAndroid.check(PermissionsAndroid.PERMISSIONS.RECORD_AUDIO))) {
      return Promise.reject(`Microphone permission is required to use Spokestack.
        Use PermissionsAndroid from React Native to request microphone
        permission before calling Spokestack.start() or Spokestack.activate().`);
    }

    return original();
  };
}) // Make all but one argument optional in JS-land
;
['synthesize', 'speak'].forEach(method => {
  const original = Spokestack[method];

  Spokestack[method] = (input, format = TTSFormat.TEXT, voice = 'demo-male') => original(input, format, voice);
}); // Ensure method is called with proper number of args despite what's passed

const originalInit = Spokestack.initialize;

Spokestack.initialize = (id, secret, config = {}) => {
  var _config$pipeline;

  // Set refreshModels to true in development by default
  if (typeof config.refreshModels !== 'boolean') {
    config.refreshModels = __DEV__;
  }
  /**
   * Wakeword
   */
  // Resolve source objects to URLs for local downloads


  const wakewordConfig = config.wakeword || {};

  if (typeof wakewordConfig.detect === 'number') {
    wakewordConfig.detect = resolveModelUrl(wakewordConfig.detect);
  }

  if (typeof wakewordConfig.encode === 'number') {
    wakewordConfig.encode = resolveModelUrl(wakewordConfig.encode);
  }

  if (typeof wakewordConfig.filter === 'number') {
    wakewordConfig.filter = resolveModelUrl(wakewordConfig.filter);
  } // Convert wakewords to a comma-separated list if an array


  if (Array.isArray(wakewordConfig.wakewords)) {
    wakewordConfig.wakewords = wakewordConfig.wakewords.join(',');
  }

  const hasWakeword = !!(wakewordConfig.filter && wakewordConfig.detect && wakewordConfig.encode);
  /**
   * Keyword
   */
  // Resolve source objects to URLs for local downloads

  const keywordConfig = config.keyword || {};

  if (typeof keywordConfig.detect === 'number') {
    keywordConfig.detect = resolveModelUrl(keywordConfig.detect);
  }

  if (typeof keywordConfig.encode === 'number') {
    keywordConfig.encode = resolveModelUrl(keywordConfig.encode);
  }

  if (typeof keywordConfig.filter === 'number') {
    keywordConfig.filter = resolveModelUrl(keywordConfig.filter);
  }

  const keywordMetadataConfig = keywordConfig;

  if (typeof keywordMetadataConfig.metadata === 'number') {
    keywordMetadataConfig.metadata = resolveModelUrl(keywordMetadataConfig.metadata);
  }

  const keywordClassesConfig = keywordConfig; // Convert classes to a comma-separated list if an array

  if (Array.isArray(keywordClassesConfig.classes)) {
    keywordClassesConfig.classes = keywordClassesConfig.classes.join(',');
  }

  const hasKeyword = !!(keywordConfig.filter && keywordConfig.detect && keywordConfig.encode && (keywordClassesConfig.classes || keywordMetadataConfig.metadata)); // Set default profile based on presence of wakeword or keyword config
  // Do not set a default if profile is set explicitly

  if (!((_config$pipeline = config.pipeline) !== null && _config$pipeline !== void 0 && _config$pipeline.profile)) {
    config.pipeline = config.pipeline || {};

    if (hasWakeword && !hasKeyword) {
      // If wakeword is set and keyword is not,
      // default to a profile that works with wakeword.
      config.pipeline.profile = PipelineProfile.TFLITE_WAKEWORD_NATIVE_ASR;
    } else if (!hasWakeword && hasKeyword) {
      // If keyword is set, and wakeword is not,
      // default to a profile that works with keyword.
      config.pipeline.profile = PipelineProfile.VAD_KEYWORD_ASR;
    } else if (hasWakeword && hasKeyword) {
      // If both keyword and wakeword are set,
      // default to a profile that works for both.
      config.pipeline.profile = PipelineProfile.TFLITE_WAKEWORD_KEYWORD;
    }
  }
  /**
   * NLU
   */


  const nluConfig = config.nlu || {};

  if (typeof nluConfig.model === 'number') {
    nluConfig.model = resolveModelUrl(nluConfig.model);
  }

  if (typeof nluConfig.metadata === 'number') {
    nluConfig.metadata = resolveModelUrl(nluConfig.metadata);
  }

  if (typeof nluConfig.vocab === 'number') {
    nluConfig.vocab = resolveModelUrl(nluConfig.vocab);
  }

  return originalInit(id, secret, config);
}; // Remove all JS listeners as well as calling
// destroy natively


const originalDestroy = Spokestack.destroy;

Spokestack.destroy = () => {
  originalDestroy(); // Remove listeners after destroy

  Spokestack.removeAllListeners();
};

Object.assign(Spokestack, {
  // Add enums as values
  PipelineProfile,
  TraceLevel,
  TTSFormat,

  /**
   * Event handling
   *
   * @example
   * const { remove } = Spokestack.addEventListener('recognize', onRecognize)
   */
  addEventListener: (type, listener) => emitter.addListener(type, listener),

  /**
   * @deprecated Just like the underlying react native method, use `remove` from the return value of `addListener`.
   */
  removeEventListener: (type, listener) => emitter.removeListener(type, listener),
  removeAllListeners: () => {
    ;
    ['recognize', 'partial_recognize', 'activate', 'deactivate', 'timeout', 'play', 'error'].forEach(event => emitter.removeAllListeners(event));
  }
}); // Export types to be used as separate exports

export * from './types'; // Allow importing these methods separately

export const {
  initialize,
  destroy,
  start,
  stop,
  activate,
  deactivate,
  synthesize,
  speak,
  classify,
  isInitialized,
  isStarted,
  isActivated,
  addEventListener,
  removeEventListener,
  removeAllListeners
} = Spokestack;
export default Spokestack;
//# sourceMappingURL=index.js.map