import { EmitterSubscription } from 'react-native';
import { PipelineProfile, SpokestackConfig, SpokestackNLUResult, TTSFormat, TraceLevel } from './types';
interface SpokestackType {
    PipelineProfile: typeof PipelineProfile;
    TraceLevel: typeof TraceLevel;
    TTSFormat: typeof TTSFormat;
    /**
     * Initialize the speech pipeline; required for all other methods.
     *
     * The first 2 args are your Spokestack credentials
     * available for free from https://spokestack.io.
     * Avoid hardcoding these in your app.
     * There are several ways to include
     * environment variables in your code.
     *
     * Using process.env:
     * https://babeljs.io/docs/en/babel-plugin-transform-inline-environment-variables/
     *
     * Using a local .env file ignored by git:
     * https://github.com/goatandsheep/react-native-dotenv
     * https://github.com/luggit/react-native-config
     *
     * See SpokestackConfig for all available options.
     *
     * @example
     * ```js
     * import Spokestack from 'react-native-spokestack'
     *
     * // ...
     *
     * await Spokestack.initialize(process.env.CLIENT_ID, process.env.CLIENT_SECRET, {
     *   pipeline: {
     *     profile: Spokestack.PipelineProfile.PTT_NATIVE_ASR
     *   }
     * })
     * ```
     */
    initialize(clientId: string, clientSecret: string, config?: SpokestackConfig): Promise<void>;
    /**
     * Destroys the speech pipeline, removes all listeners, and frees up all resources.
     * This can be called before re-initializing the pipeline.
     * A good place to call this is in `componentWillUnmount`.
     *
     * @example
     * ```js
     * componentWillUnmount() {
     *   Spokestack.destroy()
     * }
     * ```
     */
    destroy(): Promise<void>;
    /**
     * Start the speech pipeline.
     * The speech pipeline starts in the `deactivate` state.
     *
     * @example
     * ```js
     * import Spokestack from 'react-native-spokestack`
     *
     * // ...
     *
     * Spokestack.initialize(process.env.CLIENT_ID, process.env.CLIENT_SECRET)
     *   .then(Spokestack.start)
     * ```
     */
    start(): Promise<void>;
    /**
     * Stop the speech pipeline.
     * This effectively stops ASR, VAD, and wakeword.
     *
     * @example
     * ```js
     * import Spokestack from 'react-native-spokestack`
     *
     * // ...
     *
     * await Spokestack.stop()
     * ```
     */
    stop(): Promise<void>;
    /**
     * Manually activate the speech pipeline.
     * This is necessary when using a PTT profile.
     * VAD profiles can also activate ASR without the need
     * to call this method.
     *
     * @example
     * ```js
     * import Spokestack from 'react-native-spokestack`
     *
     * // ...
     *
     * <Button title="Listen" onClick={() => Spokestack.activate()} />
     * ```
     */
    activate(): Promise<void>;
    /**
     * Deactivate the speech pipeline.
     * If the profile includes wakeword, the pipeline will go back
     * to listening for the wakeword.
     * If VAD is active, the pipeline can reactivate without calling activate().
     *
     * @example
     * ```js
     * import Spokestack from 'react-native-spokestack`
     *
     * // ...
     *
     * <Button title="Stop listening" onClick={() => Spokestack.deactivate()} />
     * ```
     */
    deactivate(): Promise<void>;
    /**
     * Synthesize some text into speech
     * Returns `Promise<string>` with the string
     * being the URL for a playable mpeg.
     *
     * There is currently only one free voice available ("demo-male").
     * The voice can be changed if you have created a custom voice using
     * a Spokestack Maker account. See https://spokestack.io/pricing#maker.
     *
     * @example
     * ```js
     * const url = await Spokestack.synthesize('Hello world')
     * play(url)
     * ```
     */
    synthesize(input: string, format?: TTSFormat, voice?: string): Promise<string>;
    /**
     * Synthesize some text into speech
     * and then immediately play the audio through
     * the default audio system.
     * Audio session handling can get very complex and we recommend
     * using a RN library focused on audio for anything more than
     * very simple playback.
     *
     * There is currently only one free voice available ("demo-male").
     *
     * @example
     * ```js
     * await Spokestack.speak('Hello world')
     * ```
     */
    speak(input: string, format?: TTSFormat, voice?: string): Promise<void>;
    /**
     * Classify the utterance using the
     * intent/slot Natural Language Understanding model
     * passed to Spokestack.initialize().
     * See https://www.spokestack.io/docs/concepts/nlu for more info.
     *
     * @example
     * ```js
     * const result = await Spokestack.classify('hello')
     *
     * // Here's what the result might look like,
     * // depending on the NLU model
     * console.log(result.intent) // launch
     * ```
     */
    classify(utterance: string): Promise<SpokestackNLUResult>;
    /**
     * Returns whether Spokestack has been initialized
     *
     * @example
     * ```js
     * console.log(`isInitialized: ${await Spokestack.isInitialized()}`)
     * ```
     */
    isInitialized(): Promise<boolean>;
    /**
     * Returns whether the speech pipeline has been started
     *
     * @example
     * ```js
     * console.log(`isStarted: ${await Spokestack.isStarted()}`)
     * ```
     */
    isStarted(): Promise<boolean>;
    /**
     * Returns whether the speech pipeline is currently activated
     *
     * @example
     * ```js
     * console.log(`isActivated: ${await Spokestack.isActivated()}`)
     * ```
     */
    isActivated(): Promise<boolean>;
    /**
     * Bind to any event emitted by the native libraries
     * See Events for a list of all available events.
     *
     * @param eventType name of the event for which we are registering listener
     * @param listener the listener function
     * @param context context of the listener
     *
     * @example
     * ```js
     * useEffect(() => {
     *   const listener = Spokestack.addEventListener('recognize', onRecognize)
     *   // Unsubsribe by calling remove when components are unmounted
     *   return () => {
     *     listener.remove()
     *   }
     * }, [])
     * ```
     */
    addEventListener(eventType: string, listener: (event: any) => void, context?: Object): EmitterSubscription;
    /**
     * Remove an event listener
     *
     * @param eventType - Name of the event to emit
     * @param listener - Function to invoke when the specified event is emitted
     *
     * @example
     * ```js
     * Spokestack.removeEventListener('recognize', onRecognize)
     * ```
     */
    removeEventListener(eventType: string, listener: (...args: any[]) => any): void;
    /**
     * Remove any existing listeners
     *
     * @example
     * ```js
     * Spokestack.removeAllListeners()
     * ```
     */
    removeAllListeners(): void;
}
export * from './types';
export declare const initialize: (clientId: string, clientSecret: string, config?: SpokestackConfig | undefined) => Promise<void>, destroy: () => Promise<void>, start: () => Promise<void>, stop: () => Promise<void>, activate: () => Promise<void>, deactivate: () => Promise<void>, synthesize: (input: string, format?: TTSFormat | undefined, voice?: string | undefined) => Promise<string>, speak: (input: string, format?: TTSFormat | undefined, voice?: string | undefined) => Promise<void>, classify: (utterance: string) => Promise<SpokestackNLUResult>, isInitialized: () => Promise<boolean>, isStarted: () => Promise<boolean>, isActivated: () => Promise<boolean>, addEventListener: (eventType: string, listener: (event: any) => void, context?: Object | undefined) => EmitterSubscription, removeEventListener: (eventType: string, listener: (...args: any[]) => any) => void, removeAllListeners: () => void;
declare const _default: SpokestackType;
export default _default;
